import  IRS_toolkit.Curve as Curve
import  IRS_toolkit.legFix as legFix
import  IRS_toolkit.legFloat as legFloat
import  IRS_toolkit.Swap as Swap
from dateutil.relativedelta import relativedelta


def Price_swap(
    curve,
    Nominal,
    start_date,
    maturity_date,
    fixedRate,
    valuation_date,
    spread=0.0,
    spread1=0.0,
    ESTR=None,
    payfrecuency=None,
    Imported_schedule=None,
    type_fill="Back",
    float_start_date=None,
    legFix_date_convention="ACT/360",
    legFloat_date_convention="ACT/360",
    relative_delta=relativedelta(days=0),
):
    """This function is used to price an interest rate swap ( you need to provide correct parametrs)

    Args:
        curve (curve): yield curve
        Nominal (float): Nominal amount
        start_date (date): start date (for the fixed leg)
        maturity_date (date): maturity date
        fixedRate (float): swap fixed rate
        valuation_date (date): valuation date
        payfrecuency (int, optional): payment coupon period. Defaults to None.
        Imported_schedule (dataframe, optional): imported schedule. Defaults to None.
        type_fill (str, optional): type of filling schedule 'Backward' or 'forward'
        float_start_date (str, optional) : start date of the float leg

    Returns:
        legFix,legFloat,Swap: three main compounents of swap
    """
    if float_start_date is None:
        float_start_date = start_date

    legfix = legFix.legFix(
        Nominal,
        start_date,
        maturity_date,
        fixedRate,
        payfrecuency,
        type_fill,
        legFix_date_convention,
    )

    if Imported_schedule is not None:
        Imported_schedule["start_date"] = list(Imported_schedule["start_date"])
        Imported_schedule["end_date"] = list(Imported_schedule["end_date"])

    if Imported_schedule is not None:
        legfix.scheduleImp(Imported_schedule)
    else:
        legfix.schedule(payfrecuency)

    legfix.compute_cash_flow(valuation_date, spread, spread1)

    legfix.discount_cashflow(curve, valuation_date, relative_delta)

    # FLOAT

    legfloat = legFloat.legFloat(
        Nominal,
        float_start_date,
        maturity_date,
        curve,
        payfrecuency,
        type_fill,
        legFloat_date_convention,
        relative_delta,
    )
    if Imported_schedule is not None:
        legfloat.scheduleImp(Imported_schedule)
    else:
        legfloat.schedule(payfrecuency)
    legfloat.compute_cash_flow(valuation_date, ESTR)
    legfloat.discount_cashflow(curve, valuation_date, relative_delta)
    # SWAP
    swap = Swap.Swap(legfix, legfloat)
    return legfix, legfloat, swap


def Curves(curve_data, date_curve, shift=0.0,snowflake_instance=None):
    """This function is made to procvide you with a ZC curve  and shifted curve for a given shift

    Args:
        curve_data (dataframe): yield curve
        date_curve (date): date curve
        shift (float, optional): applied parallel shift. Defaults to 0.0.

    Returns:
        _type_: _description_
    """
    try:

        curve_data_ = curve_data[["Instrument", "Mid_geo"]]
        curve_data_.columns = ["Tenor", "Rate"]
        curve_data_.Rate = curve_data_.Rate + shift / 10000

        curve = Curve.Curve(curve_data_, date_curve,snowflake_instance)
        curve.Bootstrap()

        # up stressed curve
        curve_StressedUp = curve_data.copy()
        curve_StressedUp.Mid_geo = curve_StressedUp.Mid_geo + 0.0001
        curve_data_U = curve_StressedUp[["Instrument", "Mid_geo"]]
        curve_data_U.columns = ["Tenor", "Rate"]
        # curve ZC preprocess
        curveUp = Curve.Curve(curve_data_U, date_curve,snowflake_instance)
        curveUp.Bootstrap()

        # down curve
        curve_StressedDown = curve_data.copy()
        curve_StressedDown.Mid_geo = curve_StressedDown.Mid_geo - 0.0001
        curve_data_D = curve_StressedDown[["Instrument", "Mid_geo"]]
        curve_data_D.columns = ["Tenor", "Rate"]
        # curve ZC preprocess
        curveD = Curve.Curve(curve_data_D, date_curve,snowflake_instance)
        curveD.Bootstrap()
        return curve, curveUp, curveD
    except Exception as e:
        print("provided curve is not compatible : {e}")



import warnings
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import IRS_toolkit.utils as utils
from dateutil.relativedelta import relativedelta

warnings.filterwarnings("ignore")


class Curve:
    """
       A class for handling yield curves used in pricing and preparing
    zero-coupon curves for cash flow computation and discounting.
        Args:
               curve (dataframe): dataframe
               date_curve (date): date curve

        Attributs:
                date : curve date
                df : a dataframe that contains dates and ZC rates and Dfs

        Functions:
                setup() : create the dataframe and interpolate rate
                ForwardRate(begin,end) : compute the forward rate for two giving dates
                Bootstrap() : compute the ZC curve using the formula refer to notion appendix
                monthly_avg_daily() : compute the monthly average of instantaneous forward rates  using spot ZC curve


    """

    def __init__(self, curve=None, date_curve=None, snowflake_instance=None):
        self.date_format = "%Y-%m-%d"

        self.snowflake_instance = snowflake_instance

        if date_curve is None:
            self.date = datetime.now()
        else:
            self.date = utils.str_to_datetime(date_curve)

        self.curve = curve

        self.df = self.curve.copy()
        self.interpolated_yield_curve = None
        """    def set_periodicity(self, period):
        self.periodicity = period"""
        self.setup()

    def setup(self):
        Dates = []
        list_tenor = self.curve.iloc[:, 0].tolist()
        for tenor in list_tenor:
            Dates.append(self.date + utils.tenor_to_period(tenor))
        self.df["Date"] = Dates

        self.df.loc[-1] = ["0D", np.nan, self.date]  # adding a row
        self.df.index = self.df.index + 1  # shifting index
        self.df.sort_index(inplace=True)

        self.df.rename(columns={"Rate": "StrippedRates"}, inplace=True)
        self.df["Period"] = (self.df["Date"] - self.date).apply(lambda x: x.days)
        self.df["day_count"] = self.df.apply(
            lambda x: utils.day_count(self.date, x["Date"]), axis=1
        )

    def interpolate_yield_curve(self):
        curve_data_ = self.curve.copy()
        curve_data_["relativedelta"] = curve_data_["Tenor"].apply(
            lambda x: utils.tenor_to_period(x)
        )

        now = self.date

        curve_data_["relative_date"] = curve_data_["relativedelta"].apply(
            lambda x: (now + x)
        )
        curve_data_["Period"] = curve_data_["relative_date"].apply(
            lambda x: (x - now).days
        )

        df = pd.DataFrame({"Period": np.arange(1, max(curve_data_["Period"]) + 1, 1)})
        curve_data_ = df.merge(curve_data_, "left")
        curve_data_["Rate"] = curve_data_["Rate"].astype(float)
        curve_data_["Rate"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        curve_data_["Date"] = curve_data_["Period"].apply(lambda x: now + timedelta(x))
        curve_data_["day_count"] = curve_data_.apply(
            lambda x: utils.day_count(self.date, x["Date"]), axis=1
        )

        curve_data_.loc[-1] = [
            0,
            "0D",
            np.nan,
            relativedelta(months=0),
            self.date,
            self.date,
            0,
        ]  # adding a row
        curve_data_.index = curve_data_.index + 1  # shifting index
        curve_data_.sort_index(inplace=True)
        curve_data_.reset_index(drop=True, inplace=True)
        curve_data_.rename(columns={"Rate": "StrippedRates"}, inplace=True)
        curve_data_ = curve_data_[
            ["day_count", "StrippedRates", "Date", "Period", "Tenor"]
        ]
        self.interpolated_yield_curve = curve_data_

    def ForwardRates(
        self,
        begin,
        end,
        relative_delta=relativedelta(days=0),
        date_convention="ACT/360",
    ):
        """
        compute forward rates

        Args:
            begin (date): start date
            end (date): end date

        Returns:
            float: forward rate
        """
        try:
            # Convert string dates to datetime, if necessary
            begin_date = utils.str_to_datetime(begin) if isinstance(begin, str) else begin
            end_date = utils.str_to_datetime(end) if isinstance(end, str) else end
            end_date = end_date + relative_delta

            # Validation of date ranges
            if end_date < self.date or begin_date >= end_date:
                return None

            # Extract zero-coupon rates for the given dates
            zc_begin = self.df[self.df["Date"] == begin_date.strftime("%Y-%m-%d")]["ZC"]
            zc_end = self.df[self.df["Date"] == end_date.strftime("%Y-%m-%d")]["ZC"]

            if zc_begin.empty:
                return None  # Return None if no ZC rates found for the dates
            if zc_end.empty:
                return None  # Return None if no ZC rates found for the dates

            # Calculate discount factors (DF)
            num = (1 + zc_end.iloc[0]) ** utils.day_count(self.date, end_date)
            den = (1 + zc_begin.iloc[0]) ** utils.day_count(self.date, begin_date)
            result = (num / den) ** (
                1.0 / utils.day_count(begin_date, end_date, date_convention)
            ) - 1

            # Compute forward rate using the formula (DF2/DF1)^(1/delta(t)) - 1
            return result

        except Exception as e:
            print(f"An error occurred while calculating forward rates: {e}")
            return None  # Return a default value or None

    def Bootstrap(
        self, coupon_frequency="quarterly", date_convention="ACT/360", zc_curve=None
    ):
        """
        It Transform the yield curve to a zero-coupon (ZC) curve.

        This function processes the initial curve data to compute zero-coupon rates and discount factors.
        It handles different date calculations based on whether the current day is the first of the month.
        """
        if zc_curve is None:
            zc_curve = self.df.copy()

        coupon_periods = {
            "quarterly": 30 * 4,
            "yearly": 30,
            "monthly": 30 * 12,
            "semi_annual": 30 * 2,
        }[coupon_frequency]
        coupon_frequency_date = {
            "quarterly": 3,  # "3MS",
            "yearly": 12,
            "monthly": 1,
            "semi_annual": 6,
        }[coupon_frequency]

        zc_date = [
            self.date + relativedelta(months=i * coupon_frequency_date)
            for i in range(coupon_periods + 1)
        ]
        zc_curve_before = zc_curve[zc_curve["Date"] < zc_date[1]]
        zc_curve_before["Period"] = (zc_curve_before["Date"] - self.date).apply(
            lambda x: x.days
        )

        zc_curve_before["Coupon_period"] = zc_curve_before["day_count"]

        zc_curve_before["ZC"] = (
            1 + zc_curve_before["StrippedRates"] * zc_curve_before["Coupon_period"]
        ) ** (1 / zc_curve_before["Coupon_period"]) - 1
        zc_curve_before["DF"] = (
            1 / (1 + zc_curve_before["ZC"]) ** (zc_curve_before["Coupon_period"])
        )

        zc_curve_temp = zc_curve[zc_curve["Date"].isin(zc_date[1:])]
        zc_curve_temp.reset_index(drop=True, inplace=True)
        zc_curve_temp["Date_lagg"] = zc_curve_temp["Date"].shift()
        zc_curve_temp["Date_lagg"].fillna(self.date, inplace=True)
        zc_curve_temp["Coupon_period"] = zc_curve_temp.apply(
            lambda x: utils.day_count(x["Date_lagg"], x["Date"], date_convention), axis=1
        )
        zc_curve_temp["DF"] = 1
        for i in range(zc_curve_temp.shape[0]):
            zc_curve_temp.loc[i, "DF"] = (
                1
                - (
                    zc_curve_temp["StrippedRates"][i]
                    * zc_curve_temp["Coupon_period"]
                    * zc_curve_temp["DF"]
                )[:i].sum()
            ) / (
                1
                + zc_curve_temp["StrippedRates"][i] * zc_curve_temp["Coupon_period"][i]
            )
        zc_curve_temp["ZC"] = (1 / zc_curve_temp["DF"]) ** (
            1 / zc_curve_temp["day_count"]
        ) - 1

        zc_curve = pd.concat([zc_curve_before, zc_curve_temp[zc_curve_before.columns]])
        zc_curve.reset_index(inplace=True, drop=True)
        self.df = zc_curve.merge(zc_curve.dropna(), "left")
        dates = pd.DataFrame(
            {
                "Date": pd.date_range(
                    start=self.date,
                    end=self.date + relativedelta(years=30),
                    freq="D",
                ),
            }
        )

        self.df = dates.merge(zc_curve, "left")
        self.df["DF"] = self.df["DF"].astype(float)
        self.df["DF"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        self.df["Period"] = (self.df["Date"] - self.date).apply(lambda x: x.days)
        self.df["day_count"] = self.df.apply(
            lambda x: utils.day_count(self.date, x["Date"], date_convention), axis=1
        )
        self.df["ZC"] = (1 / self.df["DF"]) ** (1 / self.df["day_count"]) - 1
        self.df["StrippedRates"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        self.df.at[0, "DF"] = 1
        self.df.at[0, "Coupon_period"] = 0

    def monthly_avg_daily(
        self, start_date, end_date, frequency="D", relative_delta=relativedelta(days=0)
    ):
        """

        Args:
            start_date (date): start date
            end_date (date): end date

        Returns:
            Dataframe: Monthly average of daily forward rates
        """

        timedelta_dict = {
            "1D": relativedelta(day=1),
            "1W": relativedelta(weeks=1),
            "2W": relativedelta(weeks=2),
            "3W": relativedelta(weeks=3),
            "1M": relativedelta(months=1),
            "2M": relativedelta(months=2),
            "3M": relativedelta(months=3),
            "4M": relativedelta(months=4),
            "5M": relativedelta(months=5),
            "6M": relativedelta(months=6),
            "7M": relativedelta(months=7),
            "8M": relativedelta(months=8),
            "9M": relativedelta(months=9),
            "10M": relativedelta(months=10),
            "11M": relativedelta(months=11),
            "1Y": relativedelta(years=1),
            "15M": relativedelta(months=15),
            "18M": relativedelta(months=18),
            "21M": relativedelta(months=21),
            "2Y": relativedelta(years=2),
            "3Y": relativedelta(years=3),
            "4Y": relativedelta(years=4),
            "5Y": relativedelta(years=5),
            "6Y": relativedelta(years=6),
            "7Y": relativedelta(years=7),
            "8Y": relativedelta(years=8),
            "9Y": relativedelta(years=9),
            "10Y": relativedelta(years=10),
            "11Y": relativedelta(years=11),
            "12Y": relativedelta(years=12),
            "15Y": relativedelta(years=15),
            "20Y": relativedelta(years=20),
            "25Y": relativedelta(years=25),
            "30Y": relativedelta(years=30),
        }

        if frequency == "Between Tenor":
            date_list = []
            for tenor in timedelta_dict:
                date_forward = utils.str_to_datetime(start_date) + timedelta_dict[tenor]
                date_list.append(date_forward)
        else:
            date_list = pd.date_range(start_date, end=end_date, freq=frequency)

        foreward_df = pd.DataFrame([date_list[:-1], date_list[1:]]).T
        foreward_df.columns = ["start_date", "end_date"]
        foreward_list = []
        for i, j in zip(date_list[:-1], date_list[1:]):
            foreward_list.append(self.ForwardRates(i, j, relative_delta))
        foreward_df["foreward_ZC"] = foreward_list
        foreward_df["day_count"] = foreward_df.apply(
            lambda x: utils.day_count(x["start_date"], x["end_date"]), axis=1
        )
        foreward_df["foreward_simple"] = foreward_df.apply(
            lambda x: utils.ZC_to_simplerate(x["foreward_ZC"], x["day_count"]), axis=1
        )

        foreward_df = foreward_df.set_index("start_date")
        foreward_df.index = pd.to_datetime(foreward_df.index)

        return foreward_df.groupby(pd.Grouper(freq="M")).mean(), foreward_df

    def Bootstrap_12M_semi_yearly_coupon(self, coupon_frequency="quarterly"):
        """
        Transform the yield curve to a zero-coupon (ZC) curve.

        This function processes the initial curve data to compute zero-coupon rates and discount factors.
        It handles different date calculations based on whether the current day is the first of the month.
        """
        zc_curve = self.df.copy()

        coupon_periods = {
            "quarterly": 29 * 4,
            "yearly": 29,
            "monthly": 29 * 12,
            "semi_annual": 29 * 2,
        }[coupon_frequency]
        coupon_frequency_date = {
            "quarterly": pd.DateOffset(months=3),  # "3MS",
            "yearly": pd.DateOffset(years=1),
            "monthly": pd.DateOffset(months=1),
            "semi_annual": pd.DateOffset(months=6),
        }[coupon_frequency]

        # if self.date.day == 1:
        zc_date1 = pd.date_range(
            self.date.strftime(self.date_format),
            periods=2,
            freq=pd.DateOffset(months=6),
        )

        zc_date2 = pd.date_range(
            (self.date + pd.DateOffset(years=1)).strftime(self.date_format),
            periods=coupon_periods,
            freq=coupon_frequency_date,
        )

        zc_date = zc_date1.append(zc_date2)

        zc_curve_before = zc_curve[zc_curve["Date"] < zc_date[1]]
        zc_curve_before["Period"] = (zc_curve_before["Date"] - self.date).apply(
            lambda x: x.days
        )

        zc_curve_before["Coupon_period"] = zc_curve_before["day_count"]

        zc_curve_before["ZC"] = (
            1 + zc_curve_before["StrippedRates"] * zc_curve_before["Coupon_period"]
        ) ** (1 / zc_curve_before["Coupon_period"]) - 1
        zc_curve_before["DF"] = (
            1 / (1 + zc_curve_before["ZC"]) ** (zc_curve_before["Coupon_period"])
        )

        zc_curve_temp = zc_curve[zc_curve["Date"].isin(zc_date)]
        zc_curve_temp.reset_index(drop=True, inplace=True)
        zc_curve_temp["Date_lagg"] = zc_curve_temp["Date"].shift()
        zc_curve_temp["Date_lagg"].fillna(self.date, inplace=True)
        zc_curve_temp["Coupon_period"] = zc_curve_temp.apply(
            lambda x: utils.day_count(x["Date_lagg"], x["Date"]), axis=1
        )
        zc_curve_temp["DF"] = 1
        for i in range(zc_curve_temp.shape[0]):
            zc_curve_temp.loc[i, "DF"] = (
                1
                - (
                    zc_curve_temp["StrippedRates"][i]
                    * zc_curve_temp["Coupon_period"]
                    * zc_curve_temp["DF"]
                )[:i].sum()
            ) / (
                1
                + zc_curve_temp["StrippedRates"][i] * zc_curve_temp["Coupon_period"][i]
            )
        zc_curve_temp["ZC"] = (1 / zc_curve_temp["DF"]) ** (
            1 / zc_curve_temp["day_count"]
        ) - 1
        zc_curve = pd.concat([zc_curve_before, zc_curve_temp[zc_curve_before.columns]])
        zc_curve.reset_index(inplace=True, drop=True)
        self.df = self.df.merge(zc_curve.dropna(), "left")
        dates = pd.DataFrame(
            {
                "Date": pd.date_range(
                    start=self.date + relativedelta(days=1),
                    end=self.date + relativedelta(years=30),
                    freq="D",
                ),
            }
        )

        self.df = dates.merge(self.df, "left")
        self.df["DF"] = self.df["DF"].astype(float)
        self.df["DF"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        self.df["Period"] = (self.df["Date"] - self.date).apply(lambda x: x.days)
        self.df["day_count"] = self.df.apply(
            lambda x: utils.day_count(self.date, x["Date"]), axis=1
        )
        self.df["ZC"] = (1 / self.df["DF"]) ** (1 / self.df["day_count"]) - 1
        self.df["StrippedRates"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )



import datetime
import warnings

import pandas as pd
import  IRS_toolkit.utils as utils
from dateutil.relativedelta import relativedelta

warnings.filterwarnings("ignore")


class cash_flow:
    """
    cash flows handling
        Args:
            dates (Datframe): dates
            amounts (Dataframe): coupons
    """

    def __init__(self, dates, amounts):
        dates = [utils.str_to_datetime(dt) for dt in dates]
        self.cashflows = pd.DataFrame(
            {"cash_flow_date": dates, "cash_flow_amount": amounts}
        )

    def NPV(self, valuation_date, curve, relative_delta=relativedelta(days=0)):
        """
        Compute the Net present value

        Args:
            valuation_date (date): valuation date
            curve (curve): yield curve

        Returns:
            float: Net present value of future cash flows
        """
        self.cashflows["discount_cashflow_amounts"] = 0
        self.cashflows["DF"] = 0

        valuation_date = utils.str_to_datetime(valuation_date)
        valuation_date = valuation_date + datetime.timedelta(days=1)

        for ind, dt in self.cashflows.iterrows():
            if dt.cash_flow_date > valuation_date:
                self.cashflows.loc[ind, "DF"] = 1 / (
                    1
                    + curve.ForwardRates(
                        valuation_date, dt.cash_flow_date, relative_delta
                    )
                ) ** (utils.day_count(valuation_date, dt.cash_flow_date))

        self.cashflows["discount_cashflow_amounts"] = (
            self.cashflows.DF * self.cashflows.cash_flow_amount
        )
        self.NPV = self.cashflows["discount_cashflow_amounts"].sum()
        return self.cashflows["discount_cashflow_amounts"].sum()

    # wighted present value for bonds
    def WPV(self, valuation_date, curve, relative_delta=relativedelta(days=0)):
        """
        Weighted Net present value

        Args:
            valuation_date (date): valuation date
            curve (curve): yield curve

        Returns:
            float: time weightide Net present value of future cash flows
        """
        self.cashflows["wdiscount_cashflow_amounts"] = 0
        self.cashflows["DF"] = 0

        valuation_date = utils.str_to_datetime(valuation_date)

        for ind, dt in self.cashflows.iterrows():
            if dt.cash_flow_date > valuation_date:
                self.cashflows.loc[ind, "DF"] = 1 / (
                    1
                    + curve.ForwardRates(
                        valuation_date, dt.cash_flow_date, relative_delta
                    )
                ) ** (utils.day_count(valuation_date, dt.cash_flow_date))

        self.cashflows["wdiscount_cashflow_amounts"] = (
            self.cashflows.DF
            * self.cashflows.cash_flow_amount
            * self.cashflows.cash_flow_date.apply(
                lambda x: utils.day_count(valuation_date, x, "ACT/360")
            )
        )
        self.WPV = (self.cashflows["wdiscount_cashflow_amounts"]).sum()
        return self.cashflows["wdiscount_cashflow_amounts"].sum()



import warnings
from datetime import datetime, timedelta

import IRS_toolkit.Cash_flow as Cash_flow
import pandas as pd
import  IRS_toolkit.utils as utils
from dateutil.relativedelta import relativedelta

warnings.filterwarnings("ignore")
"""FixedRateBond
    this class compute the float leg of the swap

    legFloat(nominal,start_date,maturity_date,curve,periodicity)

    Attributes
    ----------
    date_format : str
        to set the format using for dates.
    face_value : float
        The notional amount of the bond.
    issue_date : date
        issue date of the the bond.
    maturity_date : date
        maturity date of the bond.
    curve : DataFrame
        ZCcurve that will be used for pricing
    coupon_rate : float
        the coupon rate of the bond
    coupon_frequency: str
        periodicity of payment
    cashflow : DataFrame
        cash flows details.
    NPV : float
        the net present value of the float leg.
    accurad_coupon : float
        the accurad coupon of the bond at the specified
        valuation date
    redemption_rate
        the redemption rate at maturity
    WPV
        Wighted present value (for duration)

    Methods
    -------
    schedule()
        Generates a schedule of payment dates based on
        the specified periodicity.
    import_schedule(Path)
        Imports a payment schedule from a specified file path.
    compute_cash_flow()
        Computes and populates the cash flow details based
        on the payment schedule.
    discount_cashflows(discount_curve)
        Discounts the cash flows using a provided discount curve.


eg:
    c=FixedRateBond('2023-08-01', '2026-08-01',
        1500000000, 0.035, 'ANNUAL', 1)

"""


class FixedRateBond:
    def __init__(
        self,
        issue_date,
        maturity_date,
        face_value,
        coupon_rate,
        coupon_frequency,
        redemption_rate,
    ):
        self.date_format = "%Y-%m-%d"
        self.issue_date = issue_date
        self.maturity_date = maturity_date
        self.face_value = face_value
        self.coupon_rate = coupon_rate
        self.coupon_frequency = coupon_frequency
        self.redemption_rate = redemption_rate
        self.schedule()
        self.compute_cash_flow()

    def Frequency(self):
        if self.coupon_frequency == "ANNUAL":
            return 12.0
        elif self.coupon_frequency == "SEMI-ANNUAL":
            return 6.0
        elif self.coupon_frequency == "QUARTERLY":
            return 3.0
        elif self.coupon_frequency == "MONTHLY":
            return 1.0
        else:
            raise ValueError("Frequency unknown try : for exemple QUARTERLY'")

    def schedule(self):
        delta = relativedelta(
            datetime.strptime(self.maturity_date, self.date_format),
            datetime.strptime(self.issue_date, self.date_format),
        )
        period = round(
            (delta.years * 12 + delta.months + delta.days / 31) / self.Frequency()
        )
        end = (
            pd.date_range(
                self.maturity_date,
                periods=period,
                freq=pd.offsets.MonthBegin(-self.Frequency()),
            )
            .to_series()
            .sort_index()
            .index
        )
        start = end + pd.offsets.MonthBegin(-self.Frequency())
        self.cashflow = pd.DataFrame({"start_date": start, "end_date": end})
        self.cashflow["start_date"][0] = self.issue_date
        self.cashflow["end_date"][-1] = self.maturity_date
        self.cashflow["Period"] = (
            self.cashflow.end_date - self.cashflow.start_date
        ).apply(lambda x: x.days)
        self.cashflow["day_count"] = self.cashflow.apply(
            lambda x: utils.day_count(x["start_date"], x["end_date"]), axis=1
        )

    def scheduleImp(self, Path):
        schedule_data = []
        schedule_data = pd.read_csv(Path, sep=";", header=0)
        # function to convert str to datetime

        def convert(strg):
            return datetime.strptime(strg, "%d/%m/%Y")

        schedule_data = schedule_data.applymap(convert)
        schedule_data["Period"] = (
            schedule_data.end_date - schedule_data.start_date
        ).apply(lambda x: x.days)
        self.cashflow = schedule_data

    def compute_cash_flow(self):
        self.cashflow["cashflow"] = (
            self.face_value
            * self.coupon_rate
            * self.cashflow.apply(
                lambda x: utils.day_count(x["start_date"], x["end_date"], "ACT/360"), axis=1
            )
        )

        self.cashflow["cashflow"][len(self.cashflow["cashflow"]) - 1] += (
            self.face_value * self.redemption_rate
        )

    def discount_cashflow(self, discount_curve, date_valo_=None):
        dates = [date.strftime(self.date_format) for date in self.cashflow.end_date]
        amounts = self.cashflow.cashflow.to_list()
        discount_cashflow = Cash_flow.cash_flow(dates, amounts)
        if date_valo_:
            date_valo = datetime.strptime(date_valo_, self.date_format)
            A = discount_cashflow.NPV(date_valo_, discount_curve)
            B = discount_cashflow.WPV(date_valo_, discount_curve)
            self.NPV = discount_cashflow.NPV
            last_coupon = self.cashflow[self.cashflow.start_date <= date_valo]
            self.accured_coupon = (
                utils.day_count(last_coupon.start_date.iloc[-1,], date_valo, "ACT/360")
                * self.coupon_rate
                * self.face_value
            )
            self.duration = B / A
        else:
            discount_cashflow.NPV(self.issue_date, discount_curve)
            self.NPV = discount_cashflow.NPV
            self.duration = (
                discount_cashflow.cashflows["discount_cashflow_amounts"]
                * self.cashflow.day_count.cumsum()
            ).sum() / self.NPV
            self.accured_coupon = 0.0

        self.cashflow[["discount_cashflow_amounts", "DF"]] = (
            discount_cashflow.cashflows[["discount_cashflow_amounts", "DF"]]
        )


"""Floaters
    this class compute the float leg of the swap

    legFloat(nominal,start_date,maturity_date,curve,periodicity)

    Attributes
    ----------
    date_format : str
        to set the format using for dates.
    face_value : float
        The notional amount of the bond.
    issue_date : date
        issue date of the the bond.
    maturity_date : date
        maturity date of the bond.
    curve : DataFrame
        ZCcurve that will be used for pricing
    coupon_frequency: str
        periodicity of payment
    cashflow : DataFrame
        cash flows details.
    NPV : float
        the net present value of the float leg.
    accurad_coupon : float
        the accurad coupon of the bond at the specified
        valuation date
    redemption_rate
        the redemption rate at maturity
    WPV
        Wighted present value (for duration)

    Methods
    -------
    schedule()
        Generates a schedule of payment dates based on
        the specified periodicity.
    import_schedule(Path)
        Imports a payment schedule from a specified file path.
    compute_cash_flow()
        Computes and populates the cash flow details based
        on the payment schedule.
    discount_cashflows(discount_curve)
        Discounts the cash flows using a provided discount curve.


eg:
    c=FloatingRateBond('2023-08-01', '2026-08-01',
        1500000000, curve, 'ANNUAL', 1)

"""


class FloatingRateBond:
    def __init__(
        self,
        issue_date,
        maturity_date,
        face_value,
        curve,
        coupon_frequency,
        redemption_rate,
    ):
        self.date_format = "%Y-%m-%d"
        self.issue_date = issue_date
        self.maturity_date = maturity_date
        self.face_value = face_value
        self.curve = curve
        self.coupon_frequency = coupon_frequency
        self.redemption_rate = redemption_rate
        self.schedule()
        self.compute_cash_flow()

    def Frequency(self):
        if self.coupon_frequency == "ANNUAL":
            return 12.0
        elif self.coupon_frequency == "SEMI-ANNUAL":
            return 6.0
        elif self.coupon_frequency == "QUARTERLY":
            return 3.0
        elif self.coupon_frequency == "MONTHLY":
            return 1.0
        else:
            raise ValueError("Frequency unknown try : for exemple 'QUARTERLY'")

    def schedule(self):
        delta = relativedelta(
            datetime.strptime(self.maturity_date, self.date_format),
            datetime.strptime(self.issue_date, self.date_format),
        )
        period = round(
            (delta.years * 12 + delta.months + delta.days / 31) / self.Frequency()
        )
        end = (
            pd.date_range(
                self.maturity_date,
                periods=period,
                freq=pd.offsets.MonthBegin(-self.Frequency()),
            )
            .to_series()
            .sort_index()
            .index
        )
        start = end + pd.offsets.MonthBegin(-self.Frequency())
        self.cashflow = pd.DataFrame({"start_date": start, "end_date": end})
        self.cashflow["start_date"][0] = self.issue_date
        self.cashflow["end_date"][-1] = self.maturity_date
        self.cashflow["Period"] = (
            self.cashflow.end_date - self.cashflow.start_date
        ).apply(lambda x: x.days)
        self.cashflow["day_count"] = self.cashflow.apply(
            lambda x: utils.day_count(x["start_date"], x["end_date"]), axis=1
        )

    def scheduleImp(self, Path):
        schedule_data = []
        schedule_data = pd.read_csv(Path, sep=";", header=0)
        # function to convert str to datetime

        def convert(strg):
            return datetime.strptime(strg, "%d/%m/%Y")

        schedule_data = schedule_data.applymap(convert)
        schedule_data["Period"] = (
            schedule_data.end_date - schedule_data.start_date
        ).apply(lambda x: x.days)
        self.cashflow = schedule_data

    def compute_cash_flow(self):
        """self.cashflow["cashflow"] = (
            self.face_value * self.coupon_rate * self.cashflow.apply(
                lambda x: day_count(x['start_date'], x["end_date"], 'ACT/360'),
                axis=1)
        )"""
        self.cashflow["forward_zc"] = self.cashflow.apply(
            lambda x: self.curve.ForwardRates(x["start_date"], x["end_date"]), axis=1
        )
        self.cashflow["forward_simple_rate"] = self.cashflow.apply(
            lambda x: utils.ZC_to_simplerate(x["forward_zc"], x["day_count"]), axis=1
        )

        self.cashflow["cashflow"] = (
            self.face_value
            * self.cashflow.forward_simple_rate
            * self.cashflow.day_count
        )

        self.cashflow["cashflow"][len(self.cashflow["cashflow"]) - 1] += (
            self.face_value * self.redemption_rate
        )

    def discount_cashflow(self, discount_curve, date_valo_=None):
        dates = [date.strftime(self.date_format) for date in self.cashflow.end_date]
        amounts = self.cashflow.cashflow.to_list()
        discount_cashflow = Cash_flow.cash_flow(dates, amounts)
        if date_valo_:
            date_valo = datetime.strptime(date_valo_, self.date_format)
            A = discount_cashflow.NPV(date_valo_, discount_curve)
            B = discount_cashflow.WPV(date_valo_, discount_curve)
            self.NPV = discount_cashflow.NPV
            self.last_coupon = self.cashflow[self.cashflow.start_date <= date_valo]
            self.next_coupon = self.cashflow[self.cashflow.start_date >= date_valo]
            # this calculation to compute the compounded ester
            ForwardCompounded = 1.0
            step = timedelta(days=1)
            date_range = [
                self.last_coupon.start_date.iloc[-1] + i * step
                for i in range(
                    (date_valo - self.last_coupon.start_date.iloc[-1]).days + 1
                )
            ]
            for i in range(len(date_range) - 1):
                ForwardCompounded *= (
                    1 + self.curve.ForwardRates(date_range[i], date_range[i + 1])
                ) ** (utils.day_count(date_range[i], date_range[i + 1], "ACT/360"))
            ForwardCompounded_ = (
                ForwardCompounded
                ** (
                    1
                    / utils.day_count(
                        self.last_coupon.start_date.iloc[-1], date_valo, "ACT/360"
                    )
                )
                - 1
            )
            self.accured_coupon = (
                utils.day_count(date_valo, self.last_coupon.end_date.iloc[-1], "ACT/360")
                * ForwardCompounded_
                * self.face_value
            )
            self.duration = B / A
        else:
            discount_cashflow.NPV(self.issue_date, discount_curve)
            self.NPV = discount_cashflow.NPV
            self.duration = (
                discount_cashflow.cashflows["discount_cashflow_amounts"]
                * self.cashflow.day_count.cumsum()
            ).sum() / self.NPV

        self.cashflow[["discount_cashflow_amounts", "DF"]] = (
            discount_cashflow.cashflows[["discount_cashflow_amounts", "DF"]]
        )



from datetime import datetime, timedelta
from math import *
from typing import Union

from dateutil.relativedelta import relativedelta
from zoneinfo import ZoneInfo

tz = ZoneInfo("Europe/Paris")

import numpy as np
import pandas as pd

############ - Function ############################

def day_count(start, end, convention="ACT/360") -> float:
    """
    This fucntion computes the period in years between two given dates
            with a defined convention
    Args:
        start (datetime): start date
        end (datetime): end date
        convention (str, optional): day count convention. Defaults to "ACT/360".

    Returns:
        float: day count with the given  day count convention
    """

    if end == start:
        day_count = 0.0
    if convention == "ACT/360":
        # year_diff = np.floor((end - start).days/365)

        # start_diff = start + pd.DateOffset(years=year_diff)

        # day_count = (end - start_diff).days/360 + year_diff
        day_count = (end - start).days / 360
        return day_count

    elif convention == "ACT/ACT":
        # to set the dates at midnight 00:00:00
        start_date = dt.datetime.combine(start, dt.datetime.min.time())
        end_date = dt.datetime.combine(end, dt.datetime.min.time())
        # is is leap or not
        year_First = 366 if calendar.isleap(start.year) else 365
        year_Last = 366 if calendar.isleap(end.year) else 365
        # count the full years
        day_count = end_date.year - start_date.year - 1.0
        # count the first and last fractions of years
        diff_firstYear = (
            dt.datetime(1 + start_date.year, 1, 1, tzinfo=tz) - start_date
        )  # tzinfo=tz
        day_count += float(diff_firstYear.days) / year_First
        diff_lastYear = end_date - dt.datetime(
            end_date.year, 1, 1, tzinfo=tz
        )  # tzinfo=tz
        day_count += float(diff_lastYear.days) / year_Last
        return day_count

    elif convention == "30/360":
        # Ensure start_date is before end_date
        if start > end:
            start, end = end, start

        # Extract year, month, day from the dates
        start_year, start_month, start_day = start.year, start.month, start.day
        end_year, end_month, end_day = end.year, end.month, end.day

        # Adjust days for 30/360 calculation
        if start_day == 31 or (
            start_month == 2 and (start_day == 29 or start_day == 28)
        ):
            start_day = 30
        if end_day == 31 and start_day == 30:
            end_day = 30

        # Calculate the difference in days
        days = (
            (end_year - start_year) * 360
            + (end_month - start_month) * 30
            + (end_day - start_day)
        ) / 360
        return days
    else:
        return 0


def ZC_to_simplerate(zc, day_count) -> float:
    """
    Convert zero coupon rate to simple rate.

    Args:
        zc (float): Zero coupon rate (compound)
        day_count (float): Period of time in years

    Returns:
        float: Simple rate
    """
    # Return 0 if day_count is 0 or zc is None to avoid division by zero or None errors
    if day_count == 0 or zc is None:
        return 0

    # Calculate and return the simple rate
    return ((1 + zc) ** day_count - 1) / day_count



def previous_coupon_date(df, valuation_date):
    """
    get the pervios coupon date with a given valuation date

    Args:
        df (Dataframe): Payments  details
        valuation_date (datetime): valuation date

    Returns:
        datetime: previous coupon date
    """

    date = valuation_date

    for ind, row in df.iterrows():
        try:
            if valuation_date >= row["start_date"] and valuation_date < row["end_date"]:
                date = row["start_date"]
        except:
            if valuation_date >= row["start date"] and valuation_date < row["end date"]:
                date = row["start date"]


    return date


def str_to_datetime(strg) -> datetime:
    """
    transfer str to datetime

    Args:
        strg (str): string date

    Returns:
        datetime: date
    """
    date_format = "%Y-%m-%d"
    if type(strg) == (str):
        return datetime.strptime(strg, date_format)
    else:
        return strg

def ExitPortion(date, amount: float) -> float:
    """_summary_

    Args:
        date (datetime): early exit date
        amount (float): amount of early exit

    Returns:
        float: early exit amount
    """
    if date < "2022-06-30":
        return -0.8 * amount / 100
    elif date < "2023-06-30":
        return -0.6 * amount / 100
    elif date < "2024-06-30":
        return -0.4 * amount / 100
    elif date < "2025-06-25":
        return -0.2 * amount / 100
    else:
        return 0.0


def PayFrequency(period: str) -> float:
    """
    coupon payment frequency

    Args:
        period (str): period

    Returns:
        float: number of month in period (default = 3)

    """
    delta = 3
    if period == "Monthly":
        delta = 1
    elif period == "Quarterly":
        delta = 3
    elif period == "Semi-Annual":
        delta = 6
    elif period == "Annual":
        delta = 12
    return delta


def Accrued_coupon(
    curve,
    Cash_flows,
    notionel,
    valuation_date,
    ESTR_df=None,
    relative_delta=relativedelta(days=0),
) -> float:
    """This function computes the accrued coupon of the float leg
        and for the past we use ESTR compounded and for the future we compute forwards

    Args:
        curve (curve): yield curve
        ESTR (dataframe): Estr compounded
        Cash_flows (Dataframe): dataframe
        notionel (float): float
        valuation_date (datetime): valuation date

    Returns:
        float: accrued coupon
    """

    if ESTR_df is not None:
        # if ESTR file is provided
        # we don't have weekends so we need to use interplation
        ESTR_df = ESTR_df.rename(
            columns={"dates": "date", "DATES": "date", "estr": "ESTR"}
        )
        ESTR = linear_interpolation(ESTR_df)
        date_min = min(ESTR["date"])
        date_max = max(ESTR["date"])
        SDate = previous_coupon_date(Cash_flows, pd.Timestamp(valuation_date))
        SDate = SDate.strftime("%Y-%m-%d")

        ESTR_start = ESTR[ESTR["date"] == SDate]["ESTR"]
        ESTR_end = ESTR[ESTR["date"] == valuation_date]["ESTR"]
        ESTR_max = ESTR[ESTR["date"] == date_max]["ESTR"]
        if (
            curve.date.strftime("%Y-%m-%d") > SDate and date_max < SDate
            # Here my start Date is a Date in which no ESTR no FORWARD RATE (can't compute the forward)
        ):
            raise ValueError(
                "Forward can't be computed (ex :Use an ESTR compounded up to curve date)"
            )

        result = 0

        if SDate < date_min or SDate > date_max:
            FRate = curve.ForwardRates(
                previous_coupon_date(Cash_flows, pd.Timestamp(valuation_date)),
                pd.Timestamp(valuation_date),
                relative_delta,
            )

            Day_count_years = day_count(
                previous_coupon_date(Cash_flows, pd.Timestamp(valuation_date)),
                pd.Timestamp(valuation_date),
            )
            Perf = 0 if FRate is None else (1 + FRate) ** Day_count_years - 1
        elif valuation_date <= date_max:
            Perf = (float(ESTR_end) / float(ESTR_start)) - 1
        elif valuation_date > date_max:
            perf_0 = (float(ESTR_max) / float(ESTR_start)) - 1
            FRate0 = curve.ForwardRates(
                pd.Timestamp(date_max) + timedelta(days=1),
                pd.Timestamp(valuation_date),
                relative_delta,
            )

            Day_count_years = day_count(
                pd.Timestamp(date_max) + timedelta(days=1),
                pd.Timestamp(valuation_date),
            )
            Perf = ((1 + FRate0) ** (Day_count_years) - 1) + perf_0 / notionel
        else:
            FRate = curve.ForwardRates(
                previous_coupon_date(Cash_flows, pd.Timestamp(valuation_date)),
                pd.Timestamp(valuation_date),
                relative_delta,
            )

            Day_count_years = day_count(
                previous_coupon_date(Cash_flows, pd.Timestamp(valuation_date)),
                pd.Timestamp(valuation_date),
            )
            Perf = 0 if FRate is None else (1 + FRate) ** Day_count_years - 1
        result = notionel * Perf
        return result

    else:
        raise ValueError("Provide an ESTR compounded xls")


def Spread_amount(cashflow, notionel, spread, valuation_date, convention="ACT/360") -> float:
    """this function compute the spread amount for a giving valuation date and start date

    Args:
        cashflow (dataframe): coupon start and end dates
        notionel (float): notionel amount
        spread (float): swap spread
        valuation_date (datetime): valuation date

    Returns:
        float: the spread amount
    """
    period = day_count(
        previous_coupon_date(cashflow, pd.Timestamp(valuation_date)),
        pd.Timestamp(valuation_date),
        convention,
    )
    return notionel * (spread) * period


def DV01(actual: float, up: float, down: float) -> float:
    """

    Args:
        actual (float): unshifted value
        up (float): value with shifted curve (+1 bps)
        down (float): value with shifted curve (-1 bps)

    Returns:
        float: sensitivity of the swap price
    """
    return (abs(actual - up) + abs(actual - down)) / 2


def linear_interpolation(df, date_column="date", value_column="ESTR"):
    """this function return a dataframe filled with the missing dates and
    calculate the linear interpolation on the values column.

    Args:
        df (DataFrame): the original DFs
        date_column (str, optional): Date. Defaults to 'date'.
        value_column (str, optional): value. Defaults to 'ESTR'.

    Returns:
        _type_: _description_
    """

    if "Date" in df.columns:
        date_column = "Date"
    elif "dates" in df.columns:
        date_column = "dates"
    elif "date" in df.columns:
        date_column = "date"
    elif "DATES" in df.columns:
        date_column = "DATES"
    df = df.sort_values(by=[date_column])

    complete_dates = pd.date_range(
        start=df[date_column].min(), end=df[date_column].max(), freq="D"
    )

    # steps
    date_column_numerical = df[date_column].apply(lambda x: x.timestamp())
    complete_dates_numerical = [d.timestamp() for d in complete_dates]

    interpolated_values = np.interp(
        complete_dates_numerical, date_column_numerical, df[value_column]
    )
    dates = [dt.strftime("%Y-%m-%d") for dt in complete_dates]
    interpolated_df = pd.DataFrame(
        {date_column: dates, value_column: interpolated_values}
    )
    return interpolated_df


def tenor_to_period(tenor: str) -> Union[timedelta, relativedelta]:
    """
    Convert a given tenor to a period.

    Args:
        tenor (str): A string representing the tenor (e.g., '1D', '2W', '3M', '1Y').

    Returns:
        Union[timedelta, relativedelta]: The corresponding period as a timedelta or relativedelta object.

    Raises:
        ValueError: If the tenor unit is invalid.

    Example:
        >>> tenor_to_period('1D')
        datetime.timedelta(days=1)
        >>> tenor_to_period('2W')
        datetime.timedelta(days=14)
        >>> tenor_to_period('3M')
        relativedelta(months=+3)
    """
    # Extract numeric value and unit from the tenor
    tenor_value = int(tenor[:-1])
    tenor_unit = tenor[-1].lower()

    # Define a dictionary mapping tenor units to their corresponding period objects
    dict_tenor = {
        'd': timedelta(days=tenor_value),
        'w': timedelta(weeks=tenor_value),
        'm': relativedelta(months=tenor_value),
        'y': relativedelta(years=tenor_value)
    }

    # Return the corresponding period if the unit is valid, otherwise raise an error
    if tenor_unit in dict_tenor:
        return dict_tenor[tenor_unit]
    else:
        raise ValueError(f"Invalid tenor unit: {tenor_unit}. Valid units are 'd', 'w', 'm', 'y'.")


def period_to_tenor(period: int) -> str:
    """
    Convert a given period in days to its corresponding tenor.

    Args:
        period (int): Number of days.

    Returns:
        str: Corresponding tenor, or None if no match is found.

    Note:
        This function assumes 30 days per month and 360 days per year.
    """
    # Ensure period is an integer
    period = int(period)

    # Define tenor dictionary with optimized calculations
    tenor_dict = {
        1: "1D", 7: "1W", 14: "2W", 21: "3W",
        **{30 * i: f"{i}M" for i in range(1, 12)},  # 1M to 11M
        360: "1Y",
        360+90: "15M", 360+180: "18M", 360+270: "21M",
        **{360 * i: f"{i}Y" for i in range(2, 13)},  # 2Y to 12Y
        360 * 15: "15Y", 360 * 20: "20Y", 360 * 25: "25Y", 360 * 30: "30Y"
    }

    # Return the tenor if found, otherwise None
    return tenor_dict.get(period)



import warnings

import  IRS_toolkit.legFix as legFix
import pandas as pd
from scipy.optimize import minimize

warnings.filterwarnings("ignore")


class Swap:
    """
    A class that provides various outputs related to swap pricing.


    Args:
        fix_leg (legFix): fixed leg
        float_leg (legFloat): float leg
    """

    def __init__(self, fix_leg, float_leg):
        self.fix_leg = fix_leg
        self.float_leg = float_leg

    def NPV(self, discount_curve, date_valo=None):
        """
        Net present value of the swap

        Args:
            discount_curve (curve): yield curve
            date_valo (date, optional): valuation date. Defaults to None.
            RunningCouv (float, optional): spread. Defaults to 0.00.
            GainGC (float, optional): spread. Defaults to 0.00.

        Returns:
            float: Net present value
        """
        self.fix_leg.discount_cashflow(discount_curve, date_valo)
        self.float_leg.discount_cashflow(discount_curve, date_valo)
        self.NPV_ = self.fix_leg.NPV - self.float_leg.NPV
        self.discount_curve = discount_curve
        # the added term here is to take in account the TRS valuation characteristics if we are valuating a simple IRS we need to put NPV(...0,0)
        # dateV = str_to_datetime(date_valo)
        return self.NPV_
        # + self.fix_leg.nominal *(RunningCouv + GainGC) * day_count(
        # previous_coupon_date(self.fix_leg.cashflow, dateV), dateV, "ACT/360")

    def fair_rate(self, date_valo, ImpSchedule=None):
        """
        fair rate of the swap

        Args:
            date_valo (date): date valuation
            ImpSchedule (dataframe) : in case you use imported schedule

        Returns:
            float, float: fair rate, theorical fair rate
        """
        fix_rate = self.fix_leg.fix_rate

        def loss_func(fix_rate):
            leg_fix = legFix.legFix(
                self.fix_leg.nominal,
                self.fix_leg.start_date,
                self.fix_leg.maturity_date,
                fix_rate,
                self.fix_leg.periodicity,
            )
            if ImpSchedule is not None:
                leg_fix.scheduleImp(ImpSchedule)
            else:
                leg_fix.schedule()
            leg_fix.compute_cash_flow(pd.Timestamp(date_valo))
            leg_fix.discount_cashflow(self.float_leg.curve, date_valo)
            return (leg_fix.NPV - self.float_leg.NPV) * (
                leg_fix.NPV - self.float_leg.NPV
            )

        res = minimize(
            loss_func,
            fix_rate,
            method="nelder-mead",
            options={"xatol": 1e-8, "disp": True},
        )
        self.faire_rate = float(res.x)
        self.faire_rate_theory = (
            self.float_leg.NPV
            / (
                self.fix_leg.nominal
                * self.fix_leg.cashflow.iloc[:, 3]
                * self.fix_leg.cashflow.DF
            ).sum()
        )
        return self.faire_rate, self.faire_rate_theory



import numpy as np
import scipy.stats as si
from scipy.optimize import fmin


def NORMSDIST(x):
    return si.norm.cdf(x, 0.0, 1.0)


class Call:
    def __init__(
        self, Stock_price=60, Strike=65, r=0.08, T=0.25, call_price=4, volatility=0.0
    ) -> None:
        self.Stock_price = Stock_price
        self.Strike = Strike
        self.r = r
        self.T = T
        self.call_price = call_price
        self.volatility = volatility

    def compute_ImpliedVolatility(self):
        def ImpliedVolatility(s):
            d1 = (
                np.log(self.Stock_price / self.Strike) + (self.r + 0.5 * s**2) * self.T
            ) / (s * np.sqrt(self.T))
            d2 = (
                np.log(self.Stock_price / self.Strike) + (self.r - 0.5 * s**2) * self.T
            ) / (s * np.sqrt(self.T))
            of = (
                self.Stock_price * NORMSDIST(d1)
                - self.Strike * np.exp(-self.r * self.T) * NORMSDIST(d2)
            ) - self.call_price
            val = of**2
            return val

        return fmin(ImpliedVolatility, 1)

    def price(self):
        d1 = (
            np.log(self.Stock_price / self.Strike)
            + (self.r + 0.5 * self.volatility**2) * self.T
        ) / (self.volatility * np.sqrt(self.T))
        d2 = (
            np.log(self.Stock_price / self.Strike)
            + (self.r - 0.5 * self.volatility**2) * self.T
        ) / (self.volatility * np.sqrt(self.T))
        BlackScholesCall = self.Stock_price * NORMSDIST(d1) - self.Strike * np.exp(
            -self.r * self.T
        ) * NORMSDIST(d2)
        self.call_price = BlackScholesCall
        return BlackScholesCall


class Put:
    def __init__(
        self,
        Stock_price=60,
        Strike=65,
        r=0.08,
        T=0.25,
        put_price=4,
        volatility=0.0,
    ) -> None:
        self.Stock_price = Stock_price
        self.Strike = Strike
        self.r = r
        self.T = T
        self.put_price = put_price
        self.volatility = volatility

    def compute_ImpliedVolatility(self):
        def ImpliedVolatility(s):
            d1 = (
                np.log(self.Stock_price / self.Strike) + (self.r + 0.5 * s**2) * self.T
            ) / (s * np.sqrt(self.T))
            d2 = (
                np.log(self.Stock_price / self.Strike) + (self.r - 0.5 * s**2) * self.T
            ) / (s * np.sqrt(self.T))
            of = (
                self.Strike * np.exp(-self.r * self.T) * NORMSDIST(-d2)
                - self.Stock_price * NORMSDIST(-d1)
            ) - self.put_price
            val = of**2
            return val

        return fmin(ImpliedVolatility, 1)

    def price(self):
        d1 = (
            np.log(self.Stock_price / self.Strike)
            + (self.r + 0.5 * self.volatility**2) * self.T
        ) / (self.volatility * np.sqrt(self.T))
        d2 = (
            np.log(self.Stock_price / self.Strike)
            + (self.r - 0.5 * self.volatility**2) * self.T
        ) / (self.volatility * np.sqrt(self.T))
        BlackScholesPut = self.Strike * np.exp(-self.r * self.T) * NORMSDIST(
            -d2
        ) - self.Stock_price * NORMSDIST(-d1)
        self.put_price = BlackScholesPut
        return BlackScholesPut



import  IRS_toolkit.Blackscholes as Black


class Swaption:
    def __init__(self):
        return

    def price(self):
        x = Black.Call(
            Stock_price=60, Strike=65, r=0.08, T=0.25, call_price=4, volatility=0.3
        )
        x.price()

        coupon_frequency = 4
        maturity = 30
        number_of_cashflows = coupon_frequency * maturity

        # A=


import math
import warnings
from datetime import datetime

import  IRS_toolkit.Cash_flow as Cash_flow
import pandas as pd
import IRS_toolkit.utils as utils
from dateutil.relativedelta import relativedelta

warnings.filterwarnings("ignore")


class legFix:
    """
    this class compute the float leg of the swap

        Args:
            nominal (float): notionel amount
            start_date (date): start date
            maturity_date (date): end date
            fix_rate (float): swap fixed rate
            periodicity (int, optional): coupon payment periodicity. Defaults to 3.
            type_fill (str, optional): type of filling schedule. Defaults to "Forward".
    """

    def __init__(
        self,
        nominal,
        start_date,
        maturity_date,
        fix_rate,
        periodicity=3,
        type_fill="Forward",
        date_convention="ACT/360",
    ):
        self.date_format = "%Y-%m-%d"
        self.nominal = nominal
        self.start_date = start_date
        self.maturity_date = maturity_date
        self.fix_rate = fix_rate
        self.periodicity = periodicity
        self.cashflow = pd.DataFrame()
        self.type_fill = type_fill
        self.accrued_coupon_fix = 0
        self.spread = 0
        self.date_convention = date_convention

    def schedule(self, freqc=3):
        """
        construct the schedule of the fixed leg

        Args:
            freqc (int, optional): frequency of coupon payment in month. Defaults to 3.
        """

        def convert(strg):
            if type(strg) == str:
                if strg == "NaT":
                    return "N/A"
                return datetime.strptime(strg, self.date_format)
            else:
                return strg

        delta = relativedelta(convert(self.maturity_date), convert(self.start_date))
        months = (delta.years * 12 + delta.months + delta.days / 30) / freqc
        period = math.ceil(months)

        if self.type_fill == "Forward":
            start_range = pd.date_range(
                start=self.start_date, periods=period, freq=pd.DateOffset(months=freqc)
            )

            end_range = start_range + pd.DateOffset(months=freqc)

            self.cashflow = pd.DataFrame(
                {"start_date": start_range, "end_date": end_range}
            )

            for i in range(1, len(self.cashflow), 1):
                if self.cashflow["start_date"][i] != self.cashflow["end_date"][i - 1]:
                    self.cashflow["start_date"][i] = self.cashflow["end_date"][i - 1]

        elif self.type_fill == "Back":
            end_range = pd.date_range(
                end=self.maturity_date, periods=period, freq=pd.DateOffset(months=freqc)
            )

            start_range = end_range - pd.DateOffset(month=freqc)

            self.cashflow = pd.DataFrame(
                {"start_date": start_range, "end_date": end_range}
            )

            for i in range(1, len(self.cashflow), 1):
                if self.cashflow["start_date"][i] != self.cashflow["end_date"][i - 1]:
                    self.cashflow["start_date"][i] = self.cashflow["end_date"][i - 1]
        else:
            raise ValueError("Error: Not an available option.")

        self.cashflow["start_date"].iloc[0] = pd.Timestamp(self.start_date)
        self.cashflow["end_date"].iloc[-1] = pd.Timestamp(self.maturity_date)

        self.cashflow["Period"] = (
            self.cashflow.end_date - self.cashflow.start_date
        ).apply(lambda x: x.days)

        self.cashflow["day_count"] = self.cashflow.apply(
            lambda x: utils.day_count(
                x["start_date"], x["end_date"], self.date_convention
            ),
            axis=1,
        )

    def scheduleImp(self, Path):
        """
        the function is an alternative of schedule function it is used to
        import the schedule

        Args:
            Path (csv): a given excel files
        """

        schedule_data = Path
        schedule_data = schedule_data.applymap(utils.str_to_datetime)
        schedule_data["Period"] = (
            schedule_data.end_date - schedule_data.start_date
        ).apply(lambda x: x.days)
        schedule_data["day_count"] = schedule_data.apply(
            lambda x: utils.day_count(
                x["start_date"], x["end_date"], self.date_convention
            ),
            axis=1,
        )

        self.cashflow = schedule_data

    def compute_cash_flow(self, date_value, spreadHC=0.0, spreadGC=0.0):
        """
        this function compute the float cash flows

        """
        self.cashflow["cashflow"] = (
            self.nominal * self.fix_rate * self.cashflow.day_count
        )
        """        for i in range(0, len(self.cashflow.end_date) - 1, 1):
            if (
                pd.Timestamp(date_value) > self.cashflow.start_date[i]
                and pd.Timestamp(date_value) < self.cashflow.end_date[i]
            ):
                self.cashflow.day_count[i] = day_count(
                    pd.Timestamp(date_value), self.cashflow.end_date[i]
                )
                self.cashflow["cashflow"][i] = (
                    self.nominal * self.fix_rate * self.cashflow.day_count[i]
                )"""

        coupon_period = utils.day_count(
            utils.previous_coupon_date(self.cashflow, pd.Timestamp(date_value)),
            pd.Timestamp(date_value),
            self.date_convention,
        )
        self.accrued_coupon_fix = self.nominal * self.fix_rate * coupon_period
        self.spreadHC = utils.Spread_amount(
            self.cashflow, self.nominal, spreadHC, date_value
        )
        self.spreadGC = utils.Spread_amount(
            self.cashflow, self.nominal, spreadGC, date_value
        )
        self.spread = self.spreadGC + self.spreadHC

    def discount_cashflow(
        self, discount_curve, date_valo=None, relative_delta=relativedelta(days=0)
    ):
        """
        used to discount future cashflows to the date_valo

        Args:
            discount_curve (curve): yield curve
            date_valo (date, optional): valuation date. Defaults to None.
        """
        dates = [date.strftime(self.date_format) for date in self.cashflow.end_date]

        amounts = self.cashflow.cashflow.to_list()

        discount_cashflow = Cash_flow.cash_flow(dates, amounts)

        if date_valo:
            discount_cashflow.NPV(date_valo, discount_curve, relative_delta)
        else:
            discount_cashflow.NPV(discount_curve.date, discount_curve, relative_delta)
        self.cashflow[["discount_cashflow_amounts", "DF"]] = (
            discount_cashflow.cashflows[["discount_cashflow_amounts", "DF"]].values
        )
        self.NPV = discount_cashflow.NPV



import pandas as pd
import  IRS_toolkit.price_swap as price_swap
import  IRS_toolkit.utils as utils


def run(
    date,
    notional_amount,
    start_date,
    maturity_date,
    valuation_date,
    fixed_rate,
    spread,
    ESTR,
    type_=["Forward", "Back"],
    periodicities=["Annual", "Semi-Annual", "Quarterly", "Monthly"],
):
    curve_date = date.strftime("%Y-%m-%d")

    curve, curveUp, curveD = price_swap.stressedcurve(curve_date)
    start_date = start_date.strftime("%Y-%m-%d")
    maturity_date = maturity_date.strftime("%Y-%m-%d")
    valuation_date = valuation_date.strftime("%Y-%m-%d")

    new_column1 = [
        "start date",
        "end date",
        "fix Period (days)",
        "fix period days",
        "fix cashflow",
        "fix DCF",
        "DF",
    ]
    new_column2 = [
        "start date",
        "end date",
        "float Period years",
        "float period years",
        "forward_ZC",
        "forward_simple_rate",
        "float cashflow",
        "float DCF",
        "float DF",
    ]
    legfix, legV, SWAP = price_swap.Price_swap(
        curve,
        notional_amount,
        start_date,
        maturity_date,
        fixed_rate / 100,
        valuation_date,
        spread / 10000,
        ESTR,
        utils.PayFrequency(periodicities),
        None,
        type_,
    )
    NPV = SWAP.NPV(curve, valuation_date)
    legfix.cashflow.columns = new_column1
    legfix.cashflow.iloc[
        :,
        [
            0,
            1,
            3,
            4,
            5,
            6,
        ],
    ]

    legV.cashflow.columns = new_column2
    legV.cashflow.iloc[
        :,
        [
            0,
            1,
            3,
            4,
            6,
            7,
        ],
    ]

    legfixU, legVU, swapU = price_swap.Price_swap(
        curveUp,
        notional_amount,
        start_date,
        maturity_date,
        fixed_rate / 100,
        valuation_date,
        spread / 10000,
        ESTR,
        utils.PayFrequency(periodicities),
        None,
        type_,
    )
    NPVU = swapU.NPV(curveUp, valuation_date)
    legfixD, legVD, swapD = price_swap.Price_swap(
        curveD,
        notional_amount,
        start_date,
        maturity_date,
        fixed_rate / 100,
        valuation_date,
        spread / 10000,
        ESTR,
        utils.PayFrequency(periodicities),
        None,
        type_,
    )
    NPVD = swapD.NPV(curveD, valuation_date)
    df_fixed = pd.DataFrame(
        {
            "NPV Fixed leg (€)": legfix.NPV,
            "Fair rate (%)": SWAP.fair_rate(valuation_date)[1] * 100,
            "Fixed Leg DV01 (€) ": utils.DV01(
                SWAP.fix_leg.NPV, swapU.fix_leg.NPV, swapD.fix_leg.NPV
            ),
        }
    )
    NPV_all_in = (
        SWAP.fix_leg.NPV
        - SWAP.float_leg.NPV
        + legfix.spread
        + legV.accrued_coupon_float
    )

    df_swap = pd.DataFrame(
        {
            "NPV of the swap (€) ": NPV,
            "Swap spread (GC & hedging costs) (€)": legfix.spread,
            "Swap DV01 (€)": utils.DV01(NPV, NPVU, NPVD),
        }
    )

    print(SWAP.float_leg.NPV, swapU.float_leg.NPV, swapD.float_leg.NPV)
    df_float = pd.DataFrame(
        {
            "NPV Float leg (€) : ": legV.NPV,
            "Accrued coupon on the float leg (€):": legV.accrued_coupon_float,
            "Float Leg DV01 (€)": utils.DV01(
                SWAP.float_leg.NPV, swapU.float_leg.NPV, swapD.float_leg.NPV
            ),
        }
    )
    return df_fixed, df_swap, df_float



import math
import warnings

import  IRS_toolkit.Cash_flow as Cash_flow
import pandas as pd
import  IRS_toolkit.utils as utils
from dateutil.relativedelta import relativedelta

warnings.filterwarnings("ignore")


class legFloat:
    """
    this class compute the float leg of the swap

        Args:
            nominal (float): notionel amount
            start_date (date): start date
            maturity_date (date): end date
            curve (curve): yield curve
            periodicity (int, optional): coupon payment periodicity. Defaults to 3.
            type_fill (str, optional): type of filling schedule. Defaults to "Forward".
    """

    def __init__(
        self,
        nominal,
        start_date,
        maturity_date,
        curve,
        periodicity=3,
        type_fill="Forward",
        date_convention="ACT/360",
        relative_delta=relativedelta(days=0),
    ):
        self.date_format = "%Y-%m-%d"
        self.nominal = nominal
        self.start_date = start_date
        self.maturity_date = maturity_date
        self.periodicity = periodicity
        self.curve = curve
        self.cashflow = pd.DataFrame()
        self.type_fill = type_fill
        self.accrued_coupon_float = 0
        self.date_convention = date_convention
        self.relative_delta = relative_delta

    def schedule(self, freqc=3):
        """
        this function construct the  schedule of float leg

        Args:
            freqc (int, optional): Payment frequency in months. Defaults to 3.

        'Raises:'
            ValueError: if type of fill is not compatible
        """
        delta = relativedelta(
            utils.str_to_datetime(self.maturity_date), utils.str_to_datetime(self.start_date)
        )

        months = (delta.years * 12 + delta.months + delta.days / 30) / freqc

        period = math.ceil(months)

        if self.type_fill == "Forward":
            start_range = pd.date_range(
                start=self.start_date, periods=period, freq=pd.DateOffset(months=freqc)
            )
            end_range = start_range + pd.DateOffset(months=freqc)

            self.cashflow = pd.DataFrame(
                {"start_date": start_range, "end_date": end_range}
            )

            for i in range(1, len(self.cashflow), 1):
                if self.cashflow["start_date"][i] != self.cashflow["end_date"][i - 1]:
                    self.cashflow["start_date"][i] = self.cashflow["end_date"][i - 1]

        elif self.type_fill == "Back":
            end_range = pd.date_range(
                end=self.maturity_date, periods=period, freq=pd.DateOffset(months=freqc)
            )
            start_range = end_range - pd.DateOffset(month=freqc)

            self.cashflow = pd.DataFrame(
                {"start_date": start_range, "end_date": end_range}
            )

            for i in range(1, len(self.cashflow), 1):
                if self.cashflow["start_date"][i] != self.cashflow["end_date"][i - 1]:
                    self.cashflow["start_date"][i] = self.cashflow["end_date"][i - 1]

        else:
            raise ValueError("Error: Not an available option.")

        self.cashflow["start_date"].iloc[0] = pd.Timestamp(self.start_date)
        self.cashflow["end_date"].iloc[-1] = pd.Timestamp(self.maturity_date)

        self.cashflow["Period"] = (
            self.cashflow.end_date - self.cashflow.start_date
        ).apply(lambda x: x.days)

        self.cashflow["day_count"] = self.cashflow.apply(
            lambda x: utils.day_count(x["start_date"], x["end_date"], self.date_convention),
            axis=1,
        )

    def scheduleImp(self, Path):
        """
        the function is an alternative of schedule function it is used to
        import the schedule

        Args:
            Path (csv): a given excel files
        """
        schedule_data = Path

        schedule_data = schedule_data.applymap(utils.str_to_datetime)

        schedule_data["Period"] = (
            schedule_data.end_date - schedule_data.start_date
        ).apply(lambda x: x.days)

        schedule_data["day_count"] = schedule_data.apply(
            lambda x: utils.day_count(x["start_date"], x["end_date"], self.date_convention),
            axis=1,
        )

        self.cashflow = schedule_data

    def compute_cash_flow(self, date_value, ESTR=None):
        """
        this function computes the float cash flows

        Args:
            date_value (date): used to split the into accrued coupon and rest
        """

        self.cashflow["forward_zc"] = self.cashflow.apply(
            lambda x: self.curve.ForwardRates(
                x["start_date"], x["end_date"], self.relative_delta
            ),
            axis=1,
        )
        # here we want to know which coupon we are computing so that we can split to float coupon and accured coupon

        for i in range(0, len(self.cashflow.end_date), 1):
            if (
                pd.Timestamp(date_value) > list(self.cashflow.start_date)[i]
                and pd.Timestamp(date_value) < list(self.cashflow.end_date)[i]
            ):
                self.cashflow["forward_zc"][self.cashflow.index[i]] = (
                    self.curve.ForwardRates(
                        pd.Timestamp(date_value),
                        list(self.cashflow.end_date)[i],
                        self.relative_delta,
                    )
                )
                self.cashflow.day_count[self.cashflow.index[i]] = utils.day_count(
                    pd.Timestamp(date_value),
                    list(self.cashflow.end_date)[i],
                    self.date_convention,
                )

        self.cashflow["forward_simple_rate"] = self.cashflow.apply(
            lambda x: utils.ZC_to_simplerate(x["forward_zc"], x["day_count"]), axis=1
        )

        self.cashflow["cashflow"] = self.nominal * (
            (1 + self.cashflow.forward_zc) ** (self.cashflow.day_count) - 1
        )

        self.accrued_coupon_float = utils.Accrued_coupon(
            self.curve, self.cashflow, self.nominal, date_value, ESTR
        )

    def discount_cashflow(
        self, discount_curve, date_valo=None, relative_delta=relativedelta(days=0)
    ):
        """
        used to discount future cashflows to the date_valo

        Args:
            discount_curve (curve): yield curve
            date_valo (date, optional): valuation date. Defaults to None.
        """
        dates = [date.strftime(self.date_format) for date in self.cashflow.end_date]

        amounts = self.cashflow.cashflow.to_list()

        discount_cashflow = Cash_flow.cash_flow(dates, amounts)

        if date_valo:
            discount_cashflow.NPV(date_valo, discount_curve, relative_delta)
        else:
            discount_cashflow.NPV(discount_curve.date, discount_curve, relative_delta)

        self.cashflow[["discount_cashflow_amounts", "DF"]] = (
            discount_cashflow.cashflows[["discount_cashflow_amounts", "DF"]].values
        )

        self.NPV = discount_cashflow.NPV + self.accrued_coupon_float
